unit bgraplanet1;

{
  Example program that use TBGRAScene3D to represent a planet image in any orientation.

  BGRAbitmap is part of LazPaint http://sourceforge.net/projects/lazpaint/

  This program is public domain.
  Original author: Patrick Chevalley pch@ap-i.net
}

{$mode objfpc}{$H+}

interface

uses BGRAScene3D, BGRABitmap, BGRABitmapTypes, BGRAMatrix3D, Math, LCLType,
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  types;

const
    vslice=64;                  // the sphere is split in 64 vertical parts
    hslice=128;                 // and 128 horizontal parts
    nslice=(vslice-1)*hslice;   // totaling 8064 face
    MoonRadius=1737.1;          // Moon radius in km
    MoonEarthDistance=384399.0; // Mean Moon-Earth distance in km
    MoonRectangularTexture='moonwac.jpg';  // size must be multiple of 2 > 128x64
    EarthRectangularTexture='earth.jpg';  // size must be multiple of 2 > 128x64

type

  { TForm1 }

  TForm1 = class(TForm)
    fpsTimer: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure FormPaint(Sender: TObject);
    procedure fpsTimerTimer(Sender: TObject);
  private
    { private declarations }
    scene: TBGRAScene3D;
    bmp: TBGRABitmap;
    moonobj: IBGRAObject3D;
    sunlight: IBGRADirectionalLight3D;
    ptex: array[0..nslice] of TBGRABitmap;
    zoom,xpos,ypos,rotalpha,rotdelta: single;
    phase,sunincl: single;
    Xref, Yref: integer;
    showfps : boolean;
    fps : single;
    prevtime: TDateTime;
    procedure Loadtexture(tfn:string);
  public
    { public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
var i: integer;
begin
  // create scene
  bmp := TBGRABitmap.Create(ClientWidth,ClientHeight,BGRABlack);
  scene := TBGRAScene3D.Create(bmp);
  scene.RenderingOptions.AntialiasingMode:=am3dResample;
  scene.RenderingOptions.AntialiasingResampleLevel:=2;
  scene.RenderingOptions.LightingInterpolation:=liLowQuality;
  scene.RenderingOptions.PerspectiveMode:=pmLinearMapping;
  scene.RenderingOptions.TextureInterpolation:=false;

  // create the texture slices
  for i:=0 to nslice do
      ptex[i] := TBGRABitmap.Create;

  // Create the Moon sphere
  moonobj:= scene.CreateSphere(MoonRadius,hslice,vslice);

  // load rectangular Moon texture
  Loadtexture(MoonRectangularTexture);

  //set ambiant lightness for a few Earthshine
  scene.AmbiantLightness := 0.005;
  //add a directional light for sunlight face
  sunlight:=scene.AddDirectionalLight(Point3D(1,0,0), 1.2);

  // Initial position and rotation
  phase:=-0.5;
  sunincl:=0;
  zoom:=1/1.1;
  xpos:=width div 2;
  ypos:=height div 2;
  rotalpha:=-pi/2;
  rotdelta:=0;
  showfps:=false;
  fps:=0;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var i: integer;
begin
  // Free scene
  scene.Free;
  bmp.Free;
  // Free face texture
  for i:=0 to nslice do
    ptex[i].Free;
end;

procedure TForm1.Loadtexture(tfn:string);
var i,tsize,nrow,ncol: integer;
    prect: Trect;
    tex: TBGRABitmap;
begin

  // load rectangular Moon texture
  tex:= TBGRABitmap.Create(tfn);

  // Initialize the texture slices
  tsize:=round(tex.Width/hslice);
  for i:=0 to nslice do
      ptex[i].SetSize(tsize,tsize);

  // Set the planet sphere
  with moonobj do begin
    // ensure unit scale
    MainPart.Scale(1);
    // Split the texture for each face
    for i:=0 to FaceCount-1 do begin
      // compute face position on full texture
      nrow := i div hslice ;
      ncol := i mod hslice ;
      prect.Top:=nrow*tsize;
      prect.Left:= ncol*tsize;
      prect.Bottom:=prect.Top+tsize;
      prect.Right:=prect.Left+tsize;
      // copy part of the texture
      ptex[i].PutImagePart(0,0,tex,prect,dmSet);
      // assign face texture
      Face[i].Texture:=ptex[i];
      // assign texture coordinates
      Face[i].TexCoord[0] := PointF(0,0);
      Face[i].TexCoord[1] := PointF(tsize-1,0);
      Face[i].TexCoord[2] := PointF(tsize-1,tsize-1);
      // polar cap only count three vertex
      if Face[i].VertexCount>3 then
         Face[i].TexCoord[3] := PointF(0,tsize-1);
    end;
  end;

  // No more need for the full texture
  tex.free;
end;

procedure TForm1.FormPaint(Sender: TObject);
var szoom: single;
    p,s,c: extended;
    sunp,viewp : TPoint3D;
    rotmatrix : TMatrix3D;
begin

  // clear image
  bmp.SetSize(ClientWidth,ClientHeight);
  bmp.FillRect(0,0,bmp.width,bmp.height,BGRABlack,dmSet);

  // rotation
  rotmatrix := MatrixRotateY(rotalpha);
  rotmatrix *= MatrixRotateX(rotdelta);

  // View from Earth distance
  viewp:=Point3D(0,0,1);
  viewp:=rotmatrix*viewp;
  scene.ViewPoint := viewp*MoonEarthDistance;

  // view position offset
  scene.ViewCenter:=PointF(xpos,ypos);

  // show Full Moon phase
  p:=phase+rotalpha;
  sincos(p,s,c);
  sunp.x:=-s;
  sunp.y:=-tan(-sunincl);
  sunp.z:=-c;
  rotmatrix := MatrixRotateX(rotdelta);
  sunp:=rotmatrix*sunp;
  Normalize3D(sunp);
  sunlight.Direction:=sunp;

  // zoom level
  szoom:=zoom*min(width,height)*(MoonEarthDistance/MoonRadius)/2;
  scene.Zoom:=PointF(szoom,szoom);

  scene.Render;

  bmp.FontHeight := 12;
  bmp.TextOut(5,5,'Right mouse : Move',ColorToBGRA(clWhite));
  bmp.TextOut(5,25,'Left mouse : Rotate',ColorToBGRA(clWhite));
  bmp.TextOut(5,45,'Mouse wheel : Zoom',ColorToBGRA(clWhite));
  bmp.TextOut(5,65,'Middle mouse : Reset',ColorToBGRA(clWhite));
  bmp.TextOut(5,85,'Keyboard left/right: Change phase',ColorToBGRA(clWhite));
  bmp.TextOut(5,105,'e: Earth',ColorToBGRA(clWhite));
  bmp.TextOut(5,125,'m: Moon',ColorToBGRA(clWhite));
  if showfps then bmp.TextOut(5,145,'fps:'+formatfloat('0.0',fps),ColorToBGRA(clWhite))
     else bmp.TextOut(5,145,'b : Benchmark',ColorToBGRA(clWhite));

  bmp.Draw(Canvas,0,0);

end;

procedure TForm1.fpsTimerTimer(Sender: TObject);
var curtime: TDateTime;
begin
  fpsTimer.Enabled:=false;
  curtime:=now;
  if prevtime>0 then begin
     fps:=1/((curtime-prevtime)*86400);
  end;
  prevtime:=curtime;
  Invalidate;
  fpsTimer.Enabled:=showfps;
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  case key of
    VK_LEFT : begin
                phase:=phase+0.1;
                Invalidate;
              end;
    VK_RIGHT: begin
                phase:=phase-0.1;
                Invalidate;
              end;
    VK_B    : begin
                showfps:=not showfps;
                fpsTimer.Enabled:=showfps;
              end;
    VK_E    : begin
                Loadtexture(EarthRectangularTexture);
                Invalidate;
              end;
    VK_M    : begin
                Loadtexture(MoonRectangularTexture);
                Invalidate;
              end;
  end;
end;

procedure TForm1.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  // store mouse position to get relative offset
  Xref:=X;
  Yref:=Y;
  // Reset position and orientation
  if Shift=[ssMiddle] then begin
    phase:=-0.5;
    sunincl:=0;
    zoom:=1/1.1;
    xpos:=width div 2;
    ypos:=height div 2;
    rotalpha:=-pi/2;
    rotdelta:=0;
    Invalidate;
  end;
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  // change position offset
  if Shift=[ssRight] then begin
    xpos:=xpos+(X-Xref);
    ypos:=ypos+(Y-Yref);
    Invalidate;
  end;
  // rotate Moon
  if Shift=[ssLeft] then begin
    rotalpha:=rotalpha-(X-Xref)/300;
    if rotalpha<0 then rotalpha:=rotalpha+2*pi;
    if rotalpha>(2*pi) then rotalpha:=rotalpha-2*pi;
    rotdelta:=rotdelta-(Y-Yref)/300;
    if rotdelta<(-pi/2) then rotdelta:=-pi/2;
    if rotdelta>(pi/2) then rotdelta:=pi/2;
    Invalidate;
  end;
  // store mouse position to get relative offset
  Xref:=X;
  Yref:=Y;
end;

procedure TForm1.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
  // mousewheel change the zoom level
  if WheelDelta>0 then
    zoom:=zoom*1.1
  else
    zoom:=zoom/1.1;
  zoom:=min(zoom,10.0);
  zoom:=max(zoom,0.1);
  Invalidate;
  Handled:=true;
end;

end.

